# Copyright 2023 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import network_mtu as mtu_def
from neutron_lib.api.definitions import provider_net
from neutron_lib import constants
from oslo_config import cfg
from oslo_utils import strutils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as ovn_config
from neutron.tests.functional import base
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.extensions import test_l3


class TestOVNClient(base.TestOVNFunctionalBase,
                    test_l3.L3NatTestCaseMixin):

    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        ext_mgr = test_l3.L3TestExtensionManager()
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)

    def test_create_metadata_port(self):
        def check_metadata_port(enable_dhcp):
            ports = self.plugin.get_ports(
                self.context, filters={'network_id': [network['id']]})
            self.assertEqual(1, len(ports))
            if enable_dhcp:
                self.assertEqual(1, len(ports[0]['fixed_ips']))
            else:
                self.assertEqual(0, len(ports[0]['fixed_ips']))
            return ports

        ovn_config.cfg.CONF.set_override('ovn_metadata_enabled', True,
                                         group='ovn')
        ovn_client = self.mech_driver._ovn_client
        for enable_dhcp in (True, False):
            network_args = {'tenant_id': 'project_1',
                            'name': 'test_net_1',
                            'admin_state_up': True,
                            'shared': False,
                            'status': constants.NET_STATUS_ACTIVE}
            network = self.plugin.create_network(self.context,
                                                 {'network': network_args})
            subnet_args = {'tenant_id': 'project_1',
                           'name': 'test_snet_1',
                           'network_id': network['id'],
                           'ip_version': constants.IP_VERSION_4,
                           'cidr': '10.210.10.0/28',
                           'enable_dhcp': enable_dhcp,
                           'gateway_ip': constants.ATTR_NOT_SPECIFIED,
                           'allocation_pools': constants.ATTR_NOT_SPECIFIED,
                           'dns_nameservers': constants.ATTR_NOT_SPECIFIED,
                           'host_routes': constants.ATTR_NOT_SPECIFIED}
            self.plugin.create_subnet(self.context, {'subnet': subnet_args})

            # The metadata port has been created during the network creation.
            ports = check_metadata_port(enable_dhcp)

            # Force the deletion and creation the metadata port.
            self.plugin.delete_port(self.context, ports[0]['id'])
            ovn_client.create_metadata_port(self.context, network)
            check_metadata_port(enable_dhcp)

            # Call again the "create_metadata_port" method as is idempotent
            # because it checks first if the metadata port exists.
            ovn_client.create_metadata_port(self.context, network)
            check_metadata_port(enable_dhcp)

    def test_create_port(self):
        with self.network('test-ovn-client') as net:
            with self.subnet(net) as subnet:
                with self.port(subnet) as port:
                    port_data = port['port']
                    nb_ovn = self.mech_driver.nb_ovn
                    lsp = nb_ovn.lsp_get(port_data['id']).execute()
                    # The logical switch port has been created during the
                    # port creation.
                    self.assertIsNotNone(lsp)
                    ovn_client = self.mech_driver._ovn_client
                    port_data = self.plugin.get_port(self.context,
                                                     port_data['id'])
                    # Call the create_port again to ensure that the create
                    # command automatically checks for existing logical
                    # switch ports
                    ovn_client.create_port(self.context, port_data)

    def test_create_router(self):
        ch = self.add_fake_chassis('host1', enable_chassis_as_gw=True,
                                   azs=[])
        net_arg = {provider_net.NETWORK_TYPE: 'geneve',
                   external_net.EXTERNAL: True}
        with self.network('test-ovn-client', as_admin=True,
                          arg_list=tuple(net_arg.keys()), **net_arg) as net:
            with self.subnet(net):
                ext_gw = {'network_id': net['network']['id']}
                with self.router(external_gateway_info=ext_gw) as router:
                    router_id = router['router']['id']
                    lr = self.nb_api.lookup('Logical_Router',
                                            ovn_utils.ovn_name(router_id))
                    self.assertEqual(ch, lr.options['chassis'])
                    lrp = lr.ports[0]
                    self.assertTrue(strutils.bool_from_string(
                        lrp.external_ids[ovn_const.OVN_ROUTER_IS_EXT_GW]))
                    hcg = self.nb_api.lookup('HA_Chassis_Group',
                                             ovn_utils.ovn_name(router_id))
                    self.assertIsNotNone(hcg)

                    # Remove the external GW port.
                    self._update('routers', router_id,
                                 {'router': {'external_gateway_info': {}}},
                                 as_admin=True)
                    lr = self.nb_api.lookup('Logical_Router',
                                            ovn_utils.ovn_name(router_id))
                    self.assertEqual([], lr.ports)
                    self.assertNotIn('chassis', lr.options)
                    hcg = self.nb_api.lookup('HA_Chassis_Group',
                                             ovn_utils.ovn_name(router_id),
                                             default=None)
                    self.assertIsNone(hcg)

    def _test_router_reside_chassis_redirect(
            self, is_distributed_fip, net_type, expected_value=None):
        cfg.CONF.set_override(
            'enable_distributed_floating_ip', is_distributed_fip, group='ovn')
        net_arg = {
            provider_net.NETWORK_TYPE: net_type}
        if net_type == constants.TYPE_FLAT:
            net_arg[provider_net.PHYSICAL_NETWORK] = 'datacentre'
        with self.network('test-ovn-client', as_admin=True,
                          arg_list=tuple(net_arg.keys()), **net_arg) as net:
            with self.subnet(net) as subnet:
                subnet_id = subnet['subnet']['id']
                with self.router() as router:
                    router_id = router['router']['id']
                    self._router_interface_action(
                        'add', router_id, subnet_id, None)
                    lr = self.nb_api.lookup('Logical_Router',
                                            ovn_utils.ovn_name(router_id))
                    lrp = lr.ports[0]
                    if net_type in [constants.TYPE_VLAN, constants.TYPE_FLAT]:
                        self.assertEqual(
                            expected_value,
                            strutils.bool_from_string(
                                lrp.options[
                                    ovn_const.LRP_OPTIONS_RESIDE_REDIR_CH]))
                    else:
                        self.assertNotIn(
                            ovn_const.LRP_OPTIONS_RESIDE_REDIR_CH,
                            lrp.options)

    def test_router_reside_chassis_redirect_dvr_vlan_net(self):
        self._test_router_reside_chassis_redirect(True, 'vlan', False)

    def test_router_reside_chassis_redirect_non_dvr_vlan_net(self):
        self._test_router_reside_chassis_redirect(False, 'vlan', True)

    def test_router_reside_chassis_redirect_dvr_flat_net(self):
        self._test_router_reside_chassis_redirect(True, 'flat', False)

    def test_router_reside_chassis_redirect_non_dvr_flat_net(self):
        self._test_router_reside_chassis_redirect(False, 'flat', True)

    def test_router_reside_chassis_redirect_dvr_geneve_net(self):
        self._test_router_reside_chassis_redirect(True, 'geneve', False)

    def test_router_reside_chassis_redirect_non_dvr_geneve_net(self):
        self._test_router_reside_chassis_redirect(False, 'geneve')

    def test_process_address_group(self):
        def _find_address_set_for_ag():
            as_v4 = self.nb_api.lookup(
                'Address_Set',
                ovn_utils.ovn_ag_addrset_name(
                    ag['id'], 'ip' + str(constants.IP_VERSION_4)),
                default=None)
            as_v6 = self.nb_api.lookup(
                'Address_Set',
                ovn_utils.ovn_ag_addrset_name(
                    ag['id'], 'ip' + str(constants.IP_VERSION_6)),
                default=None)
            return as_v4, as_v6

        ovn_client = self.mech_driver._ovn_client
        ag_args = {'project_id': 'project_1',
                   'name': 'test_address_group',
                   'description': 'test address group',
                   'addresses': ['192.168.2.2/32',
                                 '2001:db8::/32']}
        ag = self.plugin.create_address_group(self.context,
                                              {'address_group': ag_args})
        self.assertIsNotNone(_find_address_set_for_ag()[0])
        self.assertIsNotNone(_find_address_set_for_ag()[1])

        # Call the create_address_group again to ensure that the create
        # command automatically checks for existing Address_Set
        ovn_client.create_address_group(self.context, ag)

        # Update the address group
        ag['addresses'] = ['20.0.0.1/32', '2002:db8::/32']
        ovn_client.update_address_group(self.context, ag)
        as_v4_new = _find_address_set_for_ag()[0]
        as_v6_new = _find_address_set_for_ag()[1]
        self.assertEqual(['20.0.0.1/32'], as_v4_new.addresses)
        self.assertEqual(['2002:db8::/32'], as_v6_new.addresses)

        # Delete the address group
        ovn_client.delete_address_group(self.context, ag['id'])
        self.assertEqual((None, None), _find_address_set_for_ag())

    def _check_gw_lrp_mtu(self, router_id, mtu):
        # Find gateway LRP and check the MTU value.
        lr = self.nb_api.lookup('Logical_Router',
                                ovn_utils.ovn_name(router_id))
        for lrp in lr.ports:
            if strutils.bool_from_string(
                    lrp.external_ids[
                        ovn_const.OVN_ROUTER_IS_EXT_GW]):
                self.assertEqual(mtu, int(lrp.options['gateway_mtu']))
                return

        self.fail('Gateway Logical_Router_Port not found for '
                  'router %s' % router_id)

    def test_update_network_lrp_mtu_updated(self):
        self.add_fake_chassis('host1', enable_chassis_as_gw=True, azs=[])
        net_ext_args = {provider_net.NETWORK_TYPE: 'geneve',
                        external_net.EXTERNAL: True,
                        mtu_def.MTU: 1300}
        net_int_args = {provider_net.NETWORK_TYPE: 'geneve',
                        mtu_def.MTU: 1400}
        with self.network(
                'test-ext-net', as_admin=True,
                arg_list=tuple(net_ext_args.keys()), **net_ext_args) as \
                net_ext, self.network(
                'test-int-net', as_admin=True,
                arg_list=tuple(net_int_args.keys()), **net_int_args) as \
                net_int:
            with self.subnet(net_ext, cidr='10.1.0.0/24'), \
                    self.subnet(net_int, cidr='10.2.0.0/24') as snet_int:
                ext_gw = {'network_id': net_ext['network']['id']}
                with self.router(external_gateway_info=ext_gw) as router:
                    router_id = router['router']['id']
                    self._router_interface_action(
                        'add', router_id, snet_int['subnet']['id'],
                        None)

                    self._check_gw_lrp_mtu(router_id, 1300)

                    # Update external network MTU.
                    net_ext_args = {'network': {mtu_def.MTU: 1350}}
                    req = self.new_update_request('networks', net_ext_args,
                                                  net_ext['network']['id'])
                    req.get_response(self.api)
                    self._check_gw_lrp_mtu(router_id, 1350)

    def test_add_new_router_interfaces_lrp_mtu_updated(self):
        mtu_ext = 1350
        internal_interfaces = [{'cidr': '10.2.0.0/24', 'mtu': mtu_ext - 10},
                               {'cidr': '10.3.0.0/24', 'mtu': mtu_ext + 10},
                               {'cidr': '10.4.0.0/24', 'mtu': mtu_ext - 20}]
        self.add_fake_chassis('host1', enable_chassis_as_gw=True, azs=[])
        net_ext_args = {provider_net.NETWORK_TYPE: 'geneve',
                        external_net.EXTERNAL: True,
                        mtu_def.MTU: mtu_ext}
        # Store the network's MTUs connected to the router.
        router_attached_net_mtus = {mtu_ext}
        with self.network(
                'test-ext-net', as_admin=True,
                arg_list=tuple(net_ext_args.keys()), **net_ext_args) as \
                net_ext:
            ext_gw = {'network_id': net_ext['network']['id']}
            with self.subnet(net_ext, cidr='10.1.0.0/24'), \
                    self.router(external_gateway_info=ext_gw) as router:
                router_id = router['router']['id']
                for _int in internal_interfaces:
                    # Attach a new internal interface to the router. If the new
                    # network MTU is lower than the minimum MTU connected, the
                    # GW LRP.options.gateway_mtu will be updated.
                    net_int_args = {provider_net.NETWORK_TYPE: 'geneve',
                                    mtu_def.MTU: _int['mtu']}
                    router_attached_net_mtus.add(_int['mtu'])
                    net_int = self._make_network(
                        self.fmt, name='nettest', admin_state_up=True,
                        **net_int_args)
                    snet_int = self._make_subnet(
                        self.fmt, net_int, constants.ATTR_NOT_SPECIFIED,
                        _int['cidr'])
                    self._router_interface_action(
                        'add', router_id, snet_int['subnet']['id'],
                        None)

                    # The GW LRP "gateway_mtu" value should be the minimum
                    # MTU connected to the router.
                    self._check_gw_lrp_mtu(router_id,
                                           min(router_attached_net_mtus))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import copy
from unittest import mock
import uuid

import netaddr
from neutron_lib import constants
from oslo_utils import netutils
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import constants as const
from ovsdbapp import event as ovsdb_event
from ovsdbapp.tests.functional import base
from ovsdbapp.tests import utils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb \
    import impl_idl_ovn as impl
from neutron.services.portforwarding import constants as pf_const
from neutron.tests.functional import base as n_base
from neutron.tests.functional.common import ovn as ovn_common
from neutron.tests.functional.resources.ovsdb import events

OWNER = ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY
PF_PLUGIN = pf_const.PORT_FORWARDING_PLUGIN


class BaseOvnIdlTest(n_base.BaseLoggingTestCase,
                     base.FunctionalTestCase):
    schemas = ['OVN_Southbound', 'OVN_Northbound']

    def setUp(self):
        super().setUp()
        ovn_conf.register_opts()
        self.api = impl.OvsdbSbOvnIdl(self.connection['OVN_Southbound'])
        self.nbapi = impl.OvsdbNbOvnIdl(self.connection['OVN_Northbound'])
        self.handler = ovsdb_event.RowEventHandler()
        self.api.idl.notify = self.handler.notify


class TestSbApi(BaseOvnIdlTest):

    def setUp(self):
        super().setUp()
        self.data = {
            'chassis': [
                {'other_config': {'ovn-bridge-mappings':
                                  'public:br-ex,private:br-0'}},
                {'other_config': {'ovn-bridge-mappings':
                                  'public:br-ex,public2:br-ex2'}},
                {'other_config': {'ovn-bridge-mappings':
                                  'public:br-ex'}},
            ]
        }
        self.load_test_data()

    def load_test_data(self):
        with self.api.transaction(check_error=True) as txn:
            for chassis in self.data['chassis']:
                chassis['name'] = utils.get_rand_device_name('chassis')
                chassis['hostname'] = '%s.localdomain.com' % chassis['name']
                txn.add(self.api.chassis_add(
                    chassis['name'], ['geneve'], chassis['hostname'],
                    hostname=chassis['hostname'],
                    other_config=chassis['other_config']))

    def test_get_chassis_hostname_and_physnets(self):
        mapping = self.api.get_chassis_hostname_and_physnets()
        self.assertLessEqual(len(self.data['chassis']), len(mapping))
        self.assertGreaterEqual(set(mapping.keys()),
                                {c['hostname'] for c in self.data['chassis']})

    def test_get_all_chassis(self):
        chassis_list = set(self.api.get_all_chassis())
        our_chassis = {c['name'] for c in self.data['chassis']}
        self.assertLessEqual(our_chassis, chassis_list)

    def test_chassis_exists(self):
        self.assertTrue(self.api.chassis_exists(
            self.data['chassis'][0]['hostname']))
        self.assertFalse(self.api.chassis_exists("nochassishere"))

    def test_get_chassis_and_physnets(self):
        mapping = self.api.get_chassis_and_physnets()
        self.assertLessEqual(len(self.data['chassis']), len(mapping))
        self.assertGreaterEqual(set(mapping.keys()),
                                {c['name'] for c in self.data['chassis']})

    def test_multiple_physnets_in_one_bridge(self):
        self.data = {
            'chassis': [
                {'other_config': {'ovn-bridge-mappings': 'p1:br-ex,p2:br-ex'}}
            ]
        }
        self.load_test_data()
        self.assertRaises(ValueError, self.api.get_chassis_and_physnets)

    def _add_switch(self, chassis_name):
        sname = utils.get_rand_device_name(prefix='switch')
        chassis = self.api.lookup('Chassis', chassis_name)
        with self.nbapi.transaction(check_error=True) as txn:
            switch = txn.add(self.nbapi.ls_add(sname))
        return chassis, switch.result

    def _add_port_to_switch(
            self, switch, type=ovn_const.LSP_TYPE_LOCALPORT,
            device_owner=constants.DEVICE_OWNER_DISTRIBUTED):
        pname = utils.get_rand_device_name(prefix='port')
        row_event = events.WaitForCreatePortBindingEvent(pname)
        self.handler.watch_event(row_event)
        with self.nbapi.transaction(check_error=True) as txn:
            port = txn.add(self.nbapi.lsp_add(
                switch.uuid, pname, type=type,
                external_ids={
                    ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: device_owner}))
        self.assertTrue(row_event.wait())
        return port.result, row_event.row

    def test_get_metadata_port(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        result = self.api.get_metadata_port(str(binding.datapath.uuid))
        self.assertEqual(binding, result)
        self.assertEqual(binding.datapath.external_ids['logical-switch'],
                         str(switch.uuid))
        self.assertEqual(
            port.external_ids[ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY],
            constants.DEVICE_OWNER_DISTRIBUTED)

    def test_get_metadata_port_other_non_metadata_port(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        port_lbhm, binding_port_lbhm = self._add_port_to_switch(
            switch, device_owner=ovn_const.OVN_LB_HM_PORT_DISTRIBUTED)
        result = self.api.get_metadata_port(str(binding.datapath.uuid))
        self.assertEqual(binding, result)
        self.assertEqual(binding.datapath.external_ids['logical-switch'],
                         str(switch.uuid))
        self.assertEqual(
            binding_port_lbhm.datapath.external_ids['logical-switch'],
            str(switch.uuid))
        self.assertEqual(
            port_lbhm.external_ids[ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY],
            ovn_const.OVN_LB_HM_PORT_DISTRIBUTED)

    def test_get_metadata_port_missing(self):
        val = str(uuid.uuid4())
        self.assertIsNone(self.api.get_metadata_port(val))

    def _create_bound_port_with_ip(self, mac, ipaddr):
        chassis, switch = self._add_switch(
            self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        mac_ip = '{} {}'.format(mac, ipaddr)
        pb_update_event = events.WaitForUpdatePortBindingEvent(
            port.name, mac=[mac_ip])
        self.handler.watch_event(pb_update_event)
        self.nbapi.lsp_set_addresses(
            port.name, [mac_ip]).execute(check_error=True)
        self.assertTrue(pb_update_event.wait())
        self.api.lsp_bind(port.name, chassis.name).execute(check_error=True)

        return binding, switch

    def test_get_network_port_bindings_by_ip(self):
        mac = 'de:ad:be:ef:4d:ad'
        ipaddr = '192.0.2.1'
        binding, switch = self._create_bound_port_with_ip(mac, ipaddr)
        # binding, ipaddr, switch = self._create_bound_port_with_ip()
        network_id = ovn_utils.get_neutron_name(switch.name)
        result = self.api.get_network_port_bindings_by_ip(network_id, ipaddr)
        self.assertIn(binding, result)

    def test_get_network_port_bindings_by_ip_ipv6_ll(self):
        ipaddr = 'fe80::99'
        mac = str(netutils.get_mac_addr_by_ipv6(netaddr.IPAddress(ipaddr)))
        binding, switch = self._create_bound_port_with_ip(mac, ipaddr)
        network_id = ovn_utils.get_neutron_name(switch.name)
        result = self.api.get_network_port_bindings_by_ip(network_id, ipaddr)
        self.assertIn(binding, result)

    def test_get_network_port_bindings_by_ip_with_unbound_port(self):
        mac = 'de:ad:be:ef:4d:ad'
        ipaddr = '192.0.2.1'
        binding, switch = self._create_bound_port_with_ip(mac, ipaddr)
        unbound_port_name = utils.get_rand_device_name(prefix="port")
        mac_ip = "de:ad:be:ef:4d:ab %s" % ipaddr
        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(
                self.nbapi.lsp_add(switch.name, unbound_port_name, type=type))
            txn.add(self.nbapi.lsp_set_addresses(unbound_port_name, [mac_ip]))
        network_id = ovn_utils.get_neutron_name(switch.name)
        result = self.api.get_network_port_bindings_by_ip(network_id, ipaddr)
        self.assertIn(binding, result)
        self.assertEqual(1, len(result))

    def test_get_ports_on_chassis(self):
        chassis, switch = self._add_switch(
            self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        self.api.lsp_bind(port.name, chassis.name).execute(check_error=True)
        self.assertEqual([binding],
                         self.api.get_ports_on_chassis(chassis.name))

    def _test_get_ports_on_chassis_with_additional_chassis(
            self, ports, chassis, bindings, expected):
        self.api.lsp_bind(
            ports[0].name, chassis[0].name).execute(check_error=True)
        self.api.lsp_bind(
            ports[1].name, chassis[1].name).execute(check_error=True)

        self.api.db_set('Port_Binding', bindings[1].uuid,
                        additional_chassis=[chassis[0].uuid]).execute(
                check_error=True, log_errors=True)

        result = self.api.get_ports_on_chassis(
            chassis[0].name, include_additional_chassis=True)

        self.assertEqual(expected, result)

    @ovn_common.skip_if_additional_chassis_not_supported('api')
    def test_get_ports_on_chassis_with_additional_chassis(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        chassis2, switch2 = self._add_switch(self.data['chassis'][1]['name'])
        port2, binding2 = self._add_port_to_switch(switch2)

        self._test_get_ports_on_chassis_with_additional_chassis(
            ports=[port, port2],
            chassis=[chassis, chassis2],
            bindings=[binding, binding2],
            expected=[binding, binding2])

    def test_get_ports_on_chassis_with_additional_chassis_not_supported(self):
        chassis, switch = self._add_switch(self.data['chassis'][0]['name'])
        port, binding = self._add_port_to_switch(switch)
        chassis2, switch2 = self._add_switch(self.data['chassis'][1]['name'])
        port2, binding2 = self._add_port_to_switch(switch2)

        with mock.patch(
                'neutron.common.ovn.utils.is_additional_chassis_supported',
                return_value=False):
            self._test_get_ports_on_chassis_with_additional_chassis(
                ports=[port, port2],
                chassis=[chassis, chassis2],
                bindings=[binding, binding2],
                expected=[binding])


class TestNbApi(BaseOvnIdlTest):

    def setUp(self):
        super().setUp()
        self.data = {
            'lbs': [
                {'name': 'pf-floatingip-fip_id1-tcp',
                 'protocol': const.PROTO_TCP,
                 'vips': {"172.24.4.8:2020": ["10.0.0.10:22"],
                          "172.24.4.8:2021": ["10.0.0.11:22"]},
                 'external_ids': {
                     ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                         pf_const.PORT_FORWARDING_PLUGIN,
                     ovn_const.OVN_FIP_EXT_ID_KEY: 'fip_id1',
                     ovn_const.OVN_REV_NUM_EXT_ID_KEY: '1',
                     ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: 'neutron-rtr1_id'}},
                {'name': 'pf-floatingip-fip_id1-udp',
                 'protocol': const.PROTO_UDP,
                 'vips': {"172.24.4.8:53": ["10.0.0.10:53"]},
                 'external_ids': {
                     ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                         pf_const.PORT_FORWARDING_PLUGIN,
                     ovn_const.OVN_FIP_EXT_ID_KEY: 'fip_id1',
                     ovn_const.OVN_REV_NUM_EXT_ID_KEY: '1',
                     ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: 'neutron-rtr1_id'}},
                {'name': 'pf-floatingip-fip_id2-tcp',
                 'protocol': const.PROTO_TCP,
                 'vips': {"172.24.4.100:2020": ["10.0.0.10:22"]},
                 'external_ids': {
                     ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY:
                         pf_const.PORT_FORWARDING_PLUGIN,
                     ovn_const.OVN_FIP_EXT_ID_KEY: 'fip_id2',
                     ovn_const.OVN_REV_NUM_EXT_ID_KEY: '10',
                     ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY: 'neutron-rtr2_id'}},
                {'name': 'octavia_lb1_id',
                 'vips': {"172.24.4.250:80": ["10.0.0.10:8080",
                                              "10.0.0.11:8080"]},
                 'selection_fields': ['ip_dst', 'ip_src', 'tp_dst', 'tp_src'],
                 'external_ids': {
                     'enabled': 'True',
                     'lr_ref': 'neutron-rtr1_id',
                     'ls_refs': str({'neutron-net1_id': 1}),
                     ovn_const.LB_EXT_IDS_VIP_KEY: '172.24.4.250',
                     ovn_const.LB_EXT_IDS_VIP_PORT_ID_KEY: 'neutron-port1_id',
                 }},
            ],
        }
        self.load_test_data()

    def load_test_data(self):
        with self.nbapi.transaction(check_error=True) as txn:
            for lb in copy.deepcopy(self.data['lbs']):
                lb_name = lb.pop('name')
                for vip, ips in lb.pop('vips').items():
                    txn.add(self.nbapi.lb_add(
                        lb_name, vip, ips, may_exist=True, **lb))

    def test_lb_list(self):
        lbs = self.nbapi.lb_list().execute(check_error=True)
        self.assertEqual(len(self.data['lbs']), len(lbs))
        exp_values = [(lb['name'], lb['external_ids'])
                      for lb in self.data['lbs']]
        lbs_values = [(lb.name, lb.external_ids) for lb in lbs]
        self.assertCountEqual(exp_values, lbs_values)

    def test_get_router_floatingip_lbs(self):
        f = self.nbapi.get_router_floatingip_lbs
        self.assertEqual([], f('unused_router_name'))
        for exp_router_name in ['neutron-rtr1_id', 'neutron-rtr2_id']:
            exp_values = [
                (lb['name'], lb['external_ids'])
                for lb in self.data['lbs']
                if all([lb['external_ids'].get(OWNER) == PF_PLUGIN,
                        lb['external_ids'].get(
                            ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY) ==
                        exp_router_name])]
            lbs_values = [(lb.name, lb.external_ids)
                          for lb in f(exp_router_name)]
            self.assertTrue(exp_values)
            self.assertCountEqual(exp_values, lbs_values)

    def test_get_floatingip_in_nat_or_lb(self):
        f = self.nbapi.get_floatingip_in_nat_or_lb
        self.assertIsNone(f('unused_fip_id'))
        for exp_fip_id in ['fip_id1', 'fip_id2']:
            exp_values = [
                (lb['name'], lb['external_ids'])
                for lb in self.data['lbs']
                if all([lb['external_ids'].get(OWNER) == PF_PLUGIN,
                        lb['external_ids'].get(
                            ovn_const.OVN_FIP_EXT_ID_KEY) == exp_fip_id])]
            # get_floatingip_in_nat_or_lb returns the first entry
            # it finds for the fip_id provided. Make sure that it
            # is present in exp_values.
            lb_match = f(exp_fip_id)
            self.assertIn((lb_match['name'], lb_match['external_ids']),
                          exp_values)

    def test_create_lswitch_port_ha_chassis_group(self):
        ls_name = uuidutils.generate_uuid()
        lsp_name = uuidutils.generate_uuid()
        hcg_name = uuidutils.generate_uuid()
        self.nbapi.ha_chassis_group_add(hcg_name).execute(check_error=True)
        hcg = self.nbapi.lookup('HA_Chassis_Group', hcg_name)
        self.nbapi.ls_add(ls_name).execute(check_error=True)
        self.nbapi.create_lswitch_port(
            lsp_name, ls_name, ha_chassis_group=hcg.uuid).execute(
            check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual(hcg.uuid, lsp.ha_chassis_group[0].uuid)

    def test_set_lswitch_port_ha_chassis_group(self):
        ls_name = uuidutils.generate_uuid()
        lsp_name = uuidutils.generate_uuid()
        self.nbapi.ls_add(ls_name).execute(check_error=True)
        self.nbapi.create_lswitch_port(lsp_name, ls_name).execute(
            check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual([], lsp.ha_chassis_group)

        # Create an HA Chassis Group register and assign to the LSP.
        hcg_name = uuidutils.generate_uuid()
        self.nbapi.ha_chassis_group_add(hcg_name).execute(check_error=True)
        hcg = self.nbapi.lookup('HA_Chassis_Group', hcg_name)
        self.nbapi.set_lswitch_port(
            lsp_name, ha_chassis_group=hcg.uuid).execute(check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual(hcg.uuid, lsp.ha_chassis_group[0].uuid)

        # Unassign the HA Chassis Group from the LSP.
        self.nbapi.set_lswitch_port(
            lsp_name, ha_chassis_group=[]).execute(check_error=True)
        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual([], lsp.ha_chassis_group)

    def test_set_router_mac_aging(self):
        name = "aging_router1"
        with self.nbapi.transaction(check_error=True) as txn:
            r = txn.add(self.nbapi.lr_add(name))
            txn.add(self.nbapi.set_router_mac_age_limit(router=name))
        self.assertEqual(r.result.options[ovn_const.LR_OPTIONS_MAC_AGE_LIMIT],
                         ovn_conf.get_ovn_mac_binding_age_threshold())

    def test_set_router_mac_aging_all(self):
        ovn_conf.cfg.CONF.set_override("mac_binding_age_threshold", 5,
                                       group="ovn")
        names = ["aging_router2", "aging_router3"]
        with self.nbapi.transaction(check_error=True) as txn:
            for name in names:
                txn.add(self.nbapi.lr_add(name))
            txn.add(self.nbapi.set_router_mac_age_limit())
        for name in names:
            r = self.nbapi.lookup("Logical_Router", name)
            self.assertEqual(r.options[ovn_const.LR_OPTIONS_MAC_AGE_LIMIT],
                             ovn_conf.get_ovn_mac_binding_age_threshold())

    def _add_static_route(self, txn, lr_name, lrp_name, **columns):
        r = txn.add(self.nbapi.lr_add(lr_name))
        if lrp_name:
            txn.add(self.nbapi.add_lrouter_port(lrp_name, lr_name))
            columns.update({'output_port': lrp_name})
        txn.add(self.nbapi.add_static_route(lr_name, **columns))

        return r

    def test_add_static_route(self):
        name = 'router_with_static_routes'
        columns = {
            'bfd': [],
            'external_ids': {'fake_eid_key': 'fake_eid_value'},
            'ip_prefix': '0.0.0.0/0',
            'nexthop': '192.0.2.1',
            'options': {'fake_option_key': 'fake_option_value'},
            'output_port': [],
            'policy': ['dst-ip'],
            'route_table': '',
        }
        with self.nbapi.transaction(check_error=True) as txn:
            r = self._add_static_route(txn, name, '', **columns)
        for route in r.result.static_routes:
            for k, v in columns.items():
                self.assertEqual(
                    getattr(route, k),
                    v)

    def _add_static_route_bfd_assert(self, r, lr_name, lrp_name, ip_prefix,
                                     nexthop):
        for route in r.result.static_routes:
            self.assertEqual(
                route.ip_prefix,
                ip_prefix)
            self.assertEqual(
                route.nexthop,
                nexthop)
            self.assertEqual(
                route.output_port[0],
                lrp_name)

            self.assertEqual(
                route.bfd[0].logical_port,
                lrp_name)
            self.assertEqual(
                route.bfd[0].dst_ip,
                nexthop)

    def test_add_static_route_bfd(self):
        lr_name = 'router_with_static_routes_and_bfd'
        lrp_name = 'lrp-' + lr_name
        ip_prefix = '0.0.0.0/0'
        nexthop = '192.0.2.1'
        with self.nbapi.transaction(check_error=True) as txn:
            r = self._add_static_route(txn, lr_name, lrp_name,
                                       ip_prefix=ip_prefix,
                                       nexthop=nexthop,
                                       maintain_bfd=True)

        self._add_static_route_bfd_assert(r, lr_name, lrp_name, ip_prefix,
                                          nexthop)

    def test_add_static_route_bfd_record_exists(self):
        lr_name = 'router_with_static_routes_and_preexisting_bfd_record'
        lrp_name = 'lrp-' + lr_name
        ip_prefix = '0.0.0.0/0'
        nexthop = '192.0.2.1'
        with self.nbapi.transaction(check_error=True) as txn:
            bfd = txn.add(self.nbapi.bfd_add(lrp_name, nexthop))
            r = self._add_static_route(txn, lr_name, lrp_name,
                                       ip_prefix=ip_prefix,
                                       nexthop=nexthop,
                                       maintain_bfd=True)

        for route in r.result.static_routes:
            self.assertEqual(
                bfd.result,
                route.bfd[0],
            )
        self._add_static_route_bfd_assert(r, lr_name, lrp_name, ip_prefix,
                                          nexthop)

    def test_add_static_route_bfd_record_exists_multiple_txn(self):
        lr_name = 'router_with_static_routes_and_preexisting_bfd_record_txn'
        lrp_name = 'lrp-' + lr_name
        ip_prefix = '0.0.0.0/0'
        nexthop = '192.0.2.1'
        with self.nbapi.transaction(check_error=True) as txn:
            bfd = txn.add(self.nbapi.bfd_add(lrp_name, nexthop))
        with self.nbapi.transaction(check_error=True) as txn:
            r = self._add_static_route(txn, lr_name, lrp_name,
                                       ip_prefix=ip_prefix,
                                       nexthop=nexthop,
                                       maintain_bfd=True)

        for route in r.result.static_routes:
            self.assertEqual(
                bfd.result,
                route.bfd[0],
            )
        self._add_static_route_bfd_assert(r, lr_name, lrp_name, ip_prefix,
                                          nexthop)

    def test_delete_lrouter_ext_gw(self):
        lr_name = 'router_with_ext_gw'
        ip_prefix = '0.0.0.0/0'
        nexthop = '192.0.2.1'
        external_ids = {ovn_const.OVN_ROUTER_IS_EXT_GW: 'True'}
        with self.nbapi.transaction(check_error=True) as txn:
            r = self._add_static_route(txn, lr_name, '',
                                       ip_prefix=ip_prefix,
                                       nexthop=nexthop,
                                       external_ids=external_ids)

        uuids = []
        for route in r.result.static_routes:
            lkp = self.nbapi.lookup("Logical_Router_Static_Route", route.uuid)
            uuids.append(lkp.uuid)
        self.assertTrue(len(uuids))

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.delete_lrouter_ext_gw(lr_name))

        for route_uuid in uuids:
            self.assertRaises(
                idlutils.RowNotFound,
                self.nbapi.lookup,
                "Logical_Router_Static_Route",
                route_uuid)

    def test_delete_lrouter_ext_gw_bfd(self):
        lr_name = 'router_with_ext_gw_bfd'
        lrp_name = 'lrp-' + lr_name
        ip_prefix = '0.0.0.0/0'
        nexthop = '192.0.2.1'
        external_ids = {ovn_const.OVN_ROUTER_IS_EXT_GW: 'True'}
        with self.nbapi.transaction(check_error=True) as txn:
            r = self._add_static_route(txn, lr_name, lrp_name,
                                       ip_prefix=ip_prefix,
                                       nexthop=nexthop,
                                       external_ids=external_ids,
                                       maintain_bfd=True)

        uuids = []
        bfd_uuids = []
        for route in r.result.static_routes:
            lkp = self.nbapi.lookup("Logical_Router_Static_Route", route.uuid)
            uuids.append(lkp.uuid)
            self.assertTrue(len(lkp.bfd))
            for bfd_rec in lkp.bfd:
                bfd_uuids.append(bfd_rec.uuid)
        self.assertTrue(len(uuids))
        self.assertTrue(len(bfd_uuids))

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.delete_lrouter_ext_gw(lr_name))

        for route_uuid in uuids:
            self.assertRaises(
                idlutils.RowNotFound,
                self.nbapi.lookup,
                "Logical_Router_Static_Route",
                route_uuid)

        for bfd_uuid in bfd_uuids:
            self.assertRaises(
                idlutils.RowNotFound,
                self.nbapi.lookup,
                "BFD",
                bfd_uuid)

    def test_set_lsp_ha_chassis_group(self):
        with self.nbapi.transaction(check_error=True) as txn:
            ls_name = uuidutils.generate_uuid()
            lsp_name = uuidutils.generate_uuid()
            hcg_name = uuidutils.generate_uuid()
            txn.add(self.nbapi.ls_add(ls_name))
            txn.add(self.nbapi.lsp_add(ls_name, lsp_name))

        with self.nbapi.transaction(check_error=True) as txn:
            hcg = self.nbapi.ha_chassis_group_add(hcg_name)
            txn.add(hcg)
            lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
            txn.add(self.nbapi.set_lswitch_port(lsp_name,
                                                ha_chassis_group=hcg))

        lsp = self.nbapi.lookup('Logical_Switch_Port', lsp_name)
        self.assertEqual(hcg.result.uuid, lsp.ha_chassis_group[0].uuid)

    def test_delete_lrouter(self):
        router_name = ovn_utils.ovn_name(uuidutils.generate_uuid())
        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.lr_add(router_name))
            txn.add(self.nbapi.ha_chassis_group_add(router_name))

        r = self.nbapi.lookup('Logical_Router', router_name)
        self.assertEqual(router_name, r.name)
        hcg = self.nbapi.lookup('HA_Chassis_Group', router_name)
        self.assertEqual(router_name, hcg.name)

        self.nbapi.lr_del(router_name).execute(check_error=True)
        self.assertIsNone(
            self.nbapi.lookup('Logical_Router', router_name, default=None))
        self.assertIsNone(
            self.nbapi.lookup('HA_Chassis_Group', router_name, default=None))

    def _assert_routes_exist(self, lr_name, expected_count):
        lr = self.nbapi.lookup('Logical_Router', lr_name)
        actual_count = len(lr.static_routes)
        self.assertEqual(actual_count, expected_count,
                         f"Expected {expected_count} routes, "
                         f"found {actual_count}.")

    def test_del_static_routes(self):
        lr_name = ovn_utils.ovn_name(uuidutils.generate_uuid())
        routes = [('0.0.0.0/0', '192.0.2.1'), ('10.0.0.0/24', '192.0.3.1')]

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.lr_add(lr_name))
            for ip_prefix, nexthop in routes:
                txn.add(self.nbapi.add_static_route(lr_name,
                                                    ip_prefix=ip_prefix,
                                                    nexthop=nexthop))

        self._assert_routes_exist(lr_name, 2)

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.delete_static_routes(lr_name, routes))

        self._assert_routes_exist(lr_name, 0)

    def test_del_no_static_routes(self):
        lr_name = ovn_utils.ovn_name(uuidutils.generate_uuid())
        routes = []

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.lr_add(lr_name))

        self._assert_routes_exist(lr_name, 0)

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.delete_static_routes(lr_name, routes))

        self._assert_routes_exist(lr_name, 0)

    def test_modify_static_route_external_ids(self):
        lr_name = 'router_with_static_routes_and_external_ids'
        columns = {
            'bfd': [],
            'external_ids': {'fake_eid_key': 'fake_eid_value'},
            'ip_prefix': '0.0.0.0/0',
            'nexthop': '192.0.2.1',
            'options': {'fake_option_key': 'fake_option_value'},
            'output_port': [],
            'policy': ['dst-ip'],
            'route_table': '',
        }
        with self.nbapi.transaction(check_error=True) as txn:
            r = self._add_static_route(txn, lr_name, '', **columns)

        # modify the external_ids
        new_ids = {'external_ids': {'fake_eid_key': 'fake_eid_value',
                                    ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'}}

        with self.nbapi.transaction(check_error=True) as txn:
            txn.add(self.nbapi.set_static_route(r.result.static_routes[0],
                                                **new_ids))

        lr = self.nbapi.lookup('Logical_Router', lr_name)

        external_ids = {'fake_eid_key': 'fake_eid_value',
                        ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'}

        self.assertEqual(external_ids, lr.static_routes[0].external_ids)


class TestIgnoreConnectionTimeout(BaseOvnIdlTest):
    @classmethod
    def create_connection(cls, schema):
        idl = connection.OvsdbIdl.from_server(cls.schema_map[schema], schema)
        return connection.Connection(idl, 0)

    def test_setUp_will_fail_if_this_is_broken(self):
        pass



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import copy
import secrets

from oslo_log import log

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf


LOG = log.getLogger(__name__)


class OVNGatewayScheduler(metaclass=abc.ABCMeta):

    def __init__(self):
        pass

    @abc.abstractmethod
    def select(self, nb_idl, sb_idl, gateway_name, candidates=None,
               existing_chassis=None, target_lrouter=None):
        """Schedule the gateway port of a router to an OVN chassis.

        Schedule the gateway router port only if it is not already
        scheduled.
        """

    def filter_existing_chassis(self, nb_idl, gw_chassis,
                                physnet, chassis_physnets,
                                existing_chassis, az_hints, chassis_with_azs):
        chassis_list = copy.copy(existing_chassis)
        for chassis_name in existing_chassis or []:
            if utils.is_gateway_chassis_invalid(chassis_name, gw_chassis,
                                                physnet, chassis_physnets,
                                                az_hints, chassis_with_azs):
                LOG.debug("Chassis %(chassis)s is invalid for scheduling "
                          "router in physnet: %(physnet)s.",
                          {'chassis': chassis_name,
                           'physnet': physnet})
                chassis_list.remove(chassis_name)
        return chassis_list

    def _schedule_gateway(self, nb_idl, sb_idl, gateway_name, candidates,
                          existing_chassis, target_lrouter):
        existing_chassis = existing_chassis or []
        candidates = candidates or []
        candidates = list(set(candidates) - set(existing_chassis))
        # If no candidates, or gateway scheduled on MAX_GW_CHASSIS nodes
        # or all candidates in existing_chassis, return existing_chassis.
        # Otherwise, if more candidates present, then schedule them.
        if existing_chassis:
            if not candidates or (
                    len(existing_chassis) == ovn_const.MAX_GW_CHASSIS):
                return existing_chassis
        if not candidates:
            LOG.warning('Gateway %s was not scheduled on any chassis, no '
                        'candidates are available', gateway_name)
            return
        chassis_count = min(
            ovn_const.MAX_GW_CHASSIS - len(existing_chassis),
            len(candidates)
        )
        # The actual binding of the gateway to a chassis via the options
        # column or gateway_chassis column in the OVN_Northbound is done
        # by the caller
        chassis = self._select_gateway_chassis(
            nb_idl, sb_idl, candidates, 1, chassis_count, target_lrouter
        )[:chassis_count]
        # priority of existing chassis is higher than candidates
        chassis = existing_chassis + chassis

        LOG.debug("Gateway %s scheduled on chassis %s",
                  gateway_name, chassis)
        return chassis

    def _reorder_by_az(self, nb_idl, sb_idl, candidates):
        chassis_selected = []
        other_chassis = []
        azs = set()

        # Check if candidates list valid
        if not candidates:
            return candidates

        chassis_with_azs = sb_idl.get_chassis_and_azs()

        # Get list of all AZs
        for chassis in candidates:
            try:
                azs.update(chassis_with_azs[chassis])
            except KeyError:
                continue

        for chassis in candidates:
            # Verify if chassis is in an AZ not already used
            # and delete AZs of chassis from list
            try:
                chassis_azs = chassis_with_azs[chassis]
                if azs.intersection(chassis_azs):
                    azs = azs.difference(chassis_azs)
                    chassis_selected += [chassis]
                else:
                    other_chassis += [chassis]
            except KeyError:
                other_chassis += [chassis]

        chassis_selected += other_chassis

        return chassis_selected

    @abc.abstractmethod
    def _select_gateway_chassis(self, nb_idl, sb_idl, candidates,
                                priority_min, priority_max, target_lrouter):
        """Choose a chassis from candidates based on a specific policy.

        Returns a list of chassis to use for scheduling. The value at
        ``ret[0]`` will be used for the chassis with ``priority_max``, the
        value at ``ret[-1]`` will be used for the chassis with ``priority_min``
        """


class OVNGatewayChanceScheduler(OVNGatewayScheduler):
    """Randomly select an chassis for a gateway port of a router"""

    def select(self, nb_idl, sb_idl, gateway_name, candidates=None,
               existing_chassis=None, target_lrouter=None):
        return self._schedule_gateway(
            nb_idl, sb_idl, gateway_name,
            candidates, existing_chassis, target_lrouter)

    def _select_gateway_chassis(self, nb_idl, sb_idl, candidates,
                                priority_min, priority_max, target_lrouter):
        candidates = copy.deepcopy(candidates)
        secrets.SystemRandom().shuffle(candidates)
        return self._reorder_by_az(nb_idl, sb_idl, candidates)


class OVNGatewayLeastLoadedScheduler(OVNGatewayScheduler):
    """Select the least loaded chassis for a gateway port of a router"""

    def select(self, nb_idl, sb_idl, gateway_name, candidates=None,
               existing_chassis=None, target_lrouter=None):
        return self._schedule_gateway(nb_idl, sb_idl, gateway_name,
                                      candidates, existing_chassis,
                                      target_lrouter)

    def _lrp_anti_affinity_score(self, nb_idl, target_lrouter, lrp_name):
        """Provide elevated score if LRP is part of LR with multiple LRPs.

        The artificial score can be added to chassis load to make it less
        likely that the chassis is selected for scheduling.

        :param nb_idl:         IDL for the OVN NB API
        :type nb_idl:          :class:`OvsdbNbOvnIdl`
        :param target_lrouter: Logical Router object
        :type target_lrouter:  :class:`RowView`
        :param lrp_name:       Name of LRP.
        :type lrp_name:        str
        :returns:              Number, if lrp_name belongs to LR with multiple
                               LRPs.
        :rtype:                int
        """
        if target_lrouter:
            lrouter_ports = getattr(target_lrouter, 'ports', set())
            if len(lrouter_ports) and nb_idl.get_lrouter_by_lrouter_port(
                    lrp_name) == target_lrouter:
                # The `MAX_GW_CHASSIS` constant here is used mostly to get a
                # multiplier that guarantees our score will outweigh natural
                # LRP priority so that when other chassis are available those
                # will be chosen rather than a chassis already hosting a LRP
                # for this LR.
                return ovn_const.MAX_GW_CHASSIS * len(target_lrouter.ports)
        return 0

    def _select_gateway_chassis(self, nb_idl, sb_idl, candidates,
                                priority_min, priority_max, target_lrouter):
        """Returns a lit of chassis from candidates ordered by priority
        (highest first). Each chassis in every priority will be selected, as it
        is the least loaded for that specific priority.
        """
        selected_chassis = []
        priorities = list(range(priority_max, priority_min - 1, -1))
        all_chassis_bindings = nb_idl.get_all_chassis_gateway_bindings(
                candidates, priorities=priorities)
        for priority in priorities:
            chassis_load = {}
            for chassis, lrps in all_chassis_bindings.items():
                if chassis in selected_chassis:
                    continue
                lrps_with_prio = 0
                anti_affinity_score = 0
                for lrp, prio in lrps:
                    if prio == priority:
                        lrps_with_prio += 1
                    anti_affinity_score += self._lrp_anti_affinity_score(
                        nb_idl, target_lrouter, lrp)
                chassis_load[chassis] = lrps_with_prio + anti_affinity_score
            if len(chassis_load) == 0:
                break
            leastload = min(chassis_load.values())
            chassis = secrets.SystemRandom().choice(
                [chassis for chassis, load in chassis_load.items()
                 if load == leastload])
            selected_chassis.append(chassis)
        return self._reorder_by_az(nb_idl, sb_idl, selected_chassis)


OVN_SCHEDULER_STR_TO_CLASS = {
    ovn_const.OVN_L3_SCHEDULER_CHANCE: OVNGatewayChanceScheduler,
    ovn_const.OVN_L3_SCHEDULER_LEASTLOADED: OVNGatewayLeastLoadedScheduler}


def get_scheduler():
    return OVN_SCHEDULER_STR_TO_CLASS[ovn_conf.get_ovn_l3_scheduler()]()



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_utils import fileutils
import testtools
import webob

from neutron.agent.linux import utils as agent_utils
from neutron.agent.metadata import proxy_base
from neutron.agent.ovn.metadata import server as agent
from neutron.conf.agent.metadata import config as meta_conf
from neutron.tests import base

OvnPortInfo = collections.namedtuple(
        'OvnPortInfo', ['external_ids', 'chassis', 'mac', 'uuid'])


class ConfFixture(config_fixture.Config):
    pass


class TestMetadataProxyHandler(base.BaseTestCase):
    fake_conf = cfg.CONF
    fake_conf_fixture = ConfFixture(fake_conf)

    def setUp(self):
        super().setUp()
        self.useFixture(self.fake_conf_fixture)
        self.log_p = mock.patch.object(proxy_base, 'LOG')
        self.log = self.log_p.start()
        self.agent_log_p = mock.patch.object(agent, 'LOG')
        self.agent_log = self.agent_log_p.start()
        self.handler = agent.MetadataProxyHandler(self.fake_conf, 'chassis1',
                                                  mock.Mock())
        self.handler._post_fork_event.set()

    def test_call(self):
        req = mock.Mock()
        with mock.patch.object(self.handler,
                               '_get_instance_and_project_id') as get_ids:
            get_ids.return_value = ('instance_id', 'project_id')
            with mock.patch.object(self.handler, '_proxy_request') as proxy:
                proxy.return_value = 'value'

                retval = self.handler(req)
                self.assertEqual(retval, 'value')

    def test_call_no_instance_match(self):
        req = mock.Mock()
        with mock.patch.object(self.handler,
                               '_get_instance_and_project_id') as get_ids:
            get_ids.return_value = None, None
            retval = self.handler(req)
            self.assertIsInstance(retval, webob.exc.HTTPNotFound)

    def test_call_internal_server_error(self):
        req = mock.Mock()
        with mock.patch.object(self.handler,
                               '_get_instance_and_project_id') as get_ids:
            get_ids.side_effect = Exception
            retval = self.handler(req)
            self.assertIsInstance(retval, webob.exc.HTTPInternalServerError)
            self.assertEqual(len(self.log.mock_calls), 2)

    def _get_instance_and_project_id_helper(self, forwarded_for, ports,
                                            mac=None, all_ports=False):
        network_id = 'the_id'
        headers = {
            'X-Forwarded-For': forwarded_for,
            'X-OVN-Network-ID': network_id
        }

        req = mock.Mock(headers=headers)

        def mock_get_network_port_bindings_by_ip(*args, **kwargs):
            if all_ports:
                return ports
            return ports.pop(0)

        self.handler.sb_idl.get_network_port_bindings_by_ip.side_effect = (
            mock_get_network_port_bindings_by_ip)

        instance_id, project_id = (
            self.handler._get_instance_and_project_id(req))

        expected = [mock.call(network_id, forwarded_for, mac=mac)]
        self.handler.sb_idl.get_network_port_bindings_by_ip.assert_has_calls(
            expected)
        return (instance_id, project_id)

    def test_get_instance_id_network_id_ipv4(self):
        forwarded_for = '192.168.1.1'
        mac = 'fa:16:3e:12:34:56'
        ovn_port = OvnPortInfo(
            external_ids={'neutron:device_id': 'device_id',
                          'neutron:project_id': 'project_id'},
            chassis=['chassis1'],
            mac=mac,
            uuid=1)
        ports = [[ovn_port]]

        self.assertEqual(
            self._get_instance_and_project_id_helper(forwarded_for, ports),
            ('device_id', 'project_id')
        )

    def test_get_instance_id_network_id_ipv6(self):
        forwarded_for = '2001:db8::1'
        mac = 'fa:16:3e:12:34:56'
        ovn_port = OvnPortInfo(
            external_ids={'neutron:device_id': 'device_id',
                          'neutron:project_id': 'project_id'},
            chassis=['chassis1'],
            mac=mac,
            uuid=1)
        ports = [[ovn_port]]

        self.assertEqual(
            self._get_instance_and_project_id_helper(forwarded_for, ports),
            ('device_id', 'project_id')
        )

    def test_get_instance_id_network_id_ipv6_ll(self):
        forwarded_for = 'fe80::99'
        # This is the EUI encoded MAC based on the IPv6 address
        forwarded_mac = '02:00:00:00:00:99'
        ovn_port = OvnPortInfo(
            external_ids={'neutron:device_id': 'device_id',
                          'neutron:project_id': 'project_id'},
            chassis=['chassis1'],
            mac=forwarded_mac,
            uuid=1)
        ports = [[ovn_port]]

        # IPv6 and link-local, the MAC will be passed
        self.assertEqual(
            self._get_instance_and_project_id_helper(forwarded_for, ports,
                                                     mac=forwarded_mac),
            ('device_id', 'project_id')
        )

    def test_get_instance_id_network_id_no_match(self):
        forwarded_for = '192.168.1.1'
        ports = [[]]

        expected = (None, None)
        observed = self._get_instance_and_project_id_helper(forwarded_for,
                                                            ports)
        self.assertEqual(expected, observed)
        self.assertEqual(len(self.agent_log.mock_calls), 1)

    def test_get_instance_id_network_id_too_many(self):
        forwarded_for = '192.168.1.1'
        mac = 'fa:16:3e:12:34:56'
        ovn_port_1 = OvnPortInfo(
            external_ids={'neutron:device_id': 'device_id',
                          'neutron:project_id': 'project_id'},
            chassis=['chassis1'],
            mac=mac,
            uuid=1)
        ovn_port_2 = OvnPortInfo(
            external_ids={'neutron:device_id': 'device_id',
                          'neutron:project_id': 'project_id'},
            chassis=['chassis2'],
            mac=mac,
            uuid=2)
        ports = [ovn_port_1, ovn_port_2]

        expected = (None, None)
        observed = self._get_instance_and_project_id_helper(forwarded_for,
                                                            ports,
                                                            all_ports=True)
        self.assertEqual(expected, observed)
        self.assertEqual(len(self.agent_log.mock_calls), 1)


class TestUnixDomainMetadataProxy(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.cfg_p = mock.patch.object(agent, 'cfg')
        self.cfg = self.cfg_p.start()
        self.cfg.CONF.metadata_proxy_socket = '/the/path'
        self.cfg.CONF.metadata_workers = 0
        self.cfg.CONF.metadata_backlog = 128
        self.cfg.CONF.metadata_proxy_socket_mode = meta_conf.USER_MODE

    @mock.patch.object(fileutils, 'ensure_tree')
    def test_init_doesnot_exists(self, ensure_dir):
        agent.UnixDomainMetadataProxy(mock.Mock(), 'chassis1')
        ensure_dir.assert_called_once_with('/the', mode=0o755)

    def test_init_exists(self):
        with mock.patch('os.path.isdir') as isdir:
            with mock.patch('os.unlink') as unlink:
                isdir.return_value = True
                agent.UnixDomainMetadataProxy(mock.Mock(), 'chassis1')
                unlink.assert_called_once_with('/the/path')

    def test_init_exists_unlink_no_file(self):
        with mock.patch('os.path.isdir') as isdir:
            with mock.patch('os.unlink') as unlink:
                with mock.patch('os.path.exists') as exists:
                    isdir.return_value = True
                    exists.return_value = False
                    unlink.side_effect = OSError

                    agent.UnixDomainMetadataProxy(mock.Mock(), 'chassis1')
                    unlink.assert_called_once_with('/the/path')

    def test_init_exists_unlink_fails_file_still_exists(self):
        with mock.patch('os.path.isdir') as isdir:
            with mock.patch('os.unlink') as unlink:
                with mock.patch('os.path.exists') as exists:
                    isdir.return_value = True
                    exists.return_value = True
                    unlink.side_effect = OSError

                    with testtools.ExpectedException(OSError):
                        agent.UnixDomainMetadataProxy(mock.Mock(), 'chassis1')
                    unlink.assert_called_once_with('/the/path')

    @mock.patch.object(agent, 'MetadataProxyHandler')
    @mock.patch.object(agent_utils, 'UnixDomainWSGIServer')
    @mock.patch.object(fileutils, 'ensure_tree')
    def test_run(self, ensure_dir, server, handler):
        p = agent.UnixDomainMetadataProxy(self.cfg.CONF, 'chassis1')
        p.run()

        ensure_dir.assert_called_once_with('/the', mode=0o755)
        server.assert_has_calls([
            mock.call('neutron-ovn-metadata-agent'),
            mock.call().start(handler.return_value,
                              '/the/path', workers=0,
                              backlog=128, mode=0o644)]
        )



# Copyright 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import os
import signal
import tempfile

from oslo_utils import fileutils
import testscenarios

from neutron.agent.common import async_process
from neutron.agent.linux import utils
from neutron.common import utils as common_utils
from neutron.privileged.agent.linux import utils as priv_utils
from neutron.tests.functional.agent.linux import test_async_process
from neutron.tests.functional import base as functional_base


class TestPIDHelpers(test_async_process.AsyncProcessTestFramework):
    def test_get_cmdline_from_pid_and_pid_invoked_with_cmdline(self):
        cmd = ['tail', '-f', self.test_file_path]
        proc = async_process.AsyncProcess(cmd)
        proc.start(block=True)
        self.addCleanup(proc.stop)

        pid = proc.pid
        self.assertEqual(cmd, utils.get_cmdline_from_pid(pid))
        self.assertTrue(utils.pid_invoked_with_cmdline(pid, cmd))
        self.assertEqual([], utils.get_cmdline_from_pid(-1))


class TestGetRootHelperChildPid(functional_base.BaseSudoTestCase):
    def _addcleanup_sleep_process(self, parent_pid):
        sleep_pid = utils.execute(
            ['ps', '--ppid', parent_pid, '-o', 'pid=']).strip()
        self.addCleanup(
            utils.execute,
            ['kill', '-9', sleep_pid],
            check_exit_code=False,
            run_as_root=True)

    def test_get_root_helper_child_pid_returns_first_child(self):
        """Test that the first child, not lowest child pid is returned.

        Test creates following process tree:
          sudo +
               |
               +--rootwrap +
                           |
                           +--bash+
                                  |
                                  +--sleep 100

        and tests that pid of `bash' command is returned.
        """

        def wait_for_sleep_is_spawned(parent_pid):
            proc_tree = utils.execute(
                ['pstree', parent_pid], check_exit_code=False)
            processes = [command.strip() for command in proc_tree.split('---')
                         if command]
            if processes:
                return 'sleep' == processes[-1]

        cmd = ['bash', '-c', '(sleep 100)']
        proc = async_process.AsyncProcess(cmd, run_as_root=True)
        proc.start()

        # root helpers spawn their child processes asynchronously, and we
        # don't want to use proc.start(block=True) as that uses
        # get_root_helper_child_pid (The method under test) internally.
        sudo_pid = proc._process.pid
        common_utils.wait_until_true(
            functools.partial(
                wait_for_sleep_is_spawned,
                sudo_pid),
            sleep=0.1)

        child_pid = utils.get_root_helper_child_pid(
            sudo_pid, cmd, run_as_root=True)
        self.assertIsNotNone(
            child_pid,
            "get_root_helper_child_pid is expected to return the pid of the "
            "bash process")
        self._addcleanup_sleep_process(child_pid)
        with open('/proc/%s/cmdline' % child_pid) as f_proc_cmdline:
            cmdline = f_proc_cmdline.readline().split('\0')[0]
        self.assertIn('bash', cmdline)


class TestFindParentPid(functional_base.BaseSudoTestCase):

    def _stop_process(self, process):
        process.stop(kill_signal=signal.SIGKILL)

    def _test_process(self, run_as_root):
        test_pid = str(os.getppid())
        cmd = ['bash', '-c', '(sleep 10)']
        proc = async_process.AsyncProcess(cmd, run_as_root=run_as_root)
        proc.start()
        self.addCleanup(self._stop_process, proc)
        common_utils.wait_until_true(lambda: proc._process.pid,
                                     sleep=0.5, timeout=10)

        bash_pid = utils.find_parent_pid(proc._process.pid)
        testcase_pid = utils.find_parent_pid(bash_pid)
        self.assertEqual(test_pid, testcase_pid)

    def test_root_process(self):
        self._test_process(run_as_root=True)

    def test_non_root_process(self):
        self._test_process(run_as_root=False)


class TestGetProcessCountByName(functional_base.BaseSudoTestCase):

    def _stop_processes(self, processes):
        for process in processes:
            process.stop(kill_signal=signal.SIGKILL)

    def test_root_process(self):
        cmd = ['sleep', '100']
        processes = []
        for _ in range(20):
            process = async_process.AsyncProcess(cmd)
            process.start()
            processes.append(process)
        for process in processes:
            common_utils.wait_until_true(lambda: process._process.pid,
                                         sleep=0.5, timeout=5)
        self.addCleanup(self._stop_processes, processes)
        number_of_sleep = utils.get_process_count_by_name('sleep')
        # NOTE(ralonsoh): other tests can spawn sleep processes too, but at
        # this point we know there are, at least, 20 "sleep" processes running.
        self.assertLessEqual(20, number_of_sleep)


class TestFindChildPids(functional_base.BaseSudoTestCase):

    def _stop_process(self, process):
        process.stop(kill_signal=signal.SIGKILL)

    def test_find_child_pids(self):
        pid = os.getppid()
        child_pids = utils.find_child_pids(pid)
        child_pids_recursive = utils.find_child_pids(pid, recursive=True)
        for _pid in child_pids:
            self.assertIn(_pid, child_pids_recursive)

        cmd = ['sleep', '100']
        process = async_process.AsyncProcess(cmd)
        process.start()
        common_utils.wait_until_true(lambda: process._process.pid,
                                     sleep=0.5, timeout=10)
        self.addCleanup(self._stop_process, process)

        child_pids_after = utils.find_child_pids(pid)
        child_pids_recursive_after = utils.find_child_pids(pid, recursive=True)
        self.assertEqual(child_pids, child_pids_after)
        for _pid in child_pids + [process.pid]:
            self.assertIn(_pid, child_pids_recursive_after)

    def test_find_non_existing_process(self):
        with open('/proc/sys/kernel/pid_max') as fd:
            pid_max = int(fd.readline().strip())
        self.assertEqual([], utils.find_child_pids(pid_max))


class ReadIfExists(testscenarios.WithScenarios,
                   functional_base.BaseSudoTestCase):
    scenarios = [
        ('root', {'run_as_root': True}),
        ('non-root', {'run_as_root': False})]

    FILE = """Test file
line 2

line 4

"""

    @classmethod
    def _write_file(cls, path='/tmp', run_as_root=False):
        content = cls.FILE.encode('ascii')
        if run_as_root:
            return priv_utils.write_to_tempfile(content, _path=path)
        return fileutils.write_to_tempfile(content, path=path)

    def test_read_if_exists(self):
        test_file_path = self._write_file(run_as_root=self.run_as_root)
        content = utils.read_if_exists(test_file_path,
                                       run_as_root=self.run_as_root)
        file = self.FILE
        self.assertEqual(file, content)

    def test_read_if_exists_no_file(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            content = utils.read_if_exists(
                os.path.join(temp_dir, 'non-existing-file'),
                run_as_root=self.run_as_root)
        self.assertEqual('', content)



# Copyright (c) 2024 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import provider_net
from neutron_lib.api.definitions import qinq as qinq_apidef
from neutron_lib.api.definitions import vlantransparent as vlan_apidef
from oslo_config import cfg
from webob import exc as web_exc

from neutron.db import qinq_db
from neutron.plugins.ml2 import plugin as ml2_plugin
from neutron.tests.common import test_db_base_plugin_v2
from neutron.tests.unit import testlib_api


class QinqExtensionTestPlugin(ml2_plugin.Ml2Plugin,
                              qinq_db.Vlanqinq_db_mixin):
    """Test plugin to mixin the VLAN transparent extensions."""

    supported_extension_aliases = [provider_net.ALIAS,
                                   qinq_apidef.ALIAS,
                                   vlan_apidef.ALIAS]


class QinqExtensionTestCase(test_db_base_plugin_v2.TestNetworksV2):
    fmt = 'json'

    def setUp(self):
        plugin = ('neutron.tests.unit.extensions.test_qinq.'
                  'QinqExtensionTestPlugin')

        cfg.CONF.set_override('network_vlan_ranges', 'datacentre',
                              group='ml2_type_vlan')
        super().setUp(plugin=plugin)

    def test_create_network_with_qinq_attr(self):
        arg_list = (
            qinq_apidef.QINQ_FIELD),
        net_kwargs = {
            qinq_apidef.QINQ_FIELD: True
        }
        with self.network(name='net1', as_admin=True,
                          arg_list=arg_list, **net_kwargs) as net:
            req = self.new_show_request('networks', net['network']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(net['network']['name'],
                             res['network']['name'])
            self.assertTrue(res['network'][qinq_apidef.QINQ_FIELD])

    def test_create_network_with_bad_qinq_attr(self):
        arg_list = (
            qinq_apidef.QINQ_FIELD),
        net_kwargs = {
            qinq_apidef.QINQ_FIELD: 'this is not boolean value',
        }
        with testlib_api.ExpectedException(
                web_exc.HTTPClientError) as ctx_manager:
            with self.network(name='net1', as_admin=True,
                              arg_list=arg_list, **net_kwargs):
                pass
        self.assertEqual(web_exc.HTTPClientError.code,
                         ctx_manager.exception.code)

    def test_network_update_with_qinq_exception(self):
        arg_list = (
            qinq_apidef.QINQ_FIELD),
        net_kwargs = {
            qinq_apidef.QINQ_FIELD: False,
        }
        with self.network(name='net1', as_admin=True,
                          arg_list=arg_list, **net_kwargs) as net:
            self._update('networks', net['network']['id'],
                         {'network': {qinq_apidef.QINQ_FIELD: True}},
                         web_exc.HTTPBadRequest.code)
            req = self.new_show_request('networks', net['network']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(net['network']['name'],
                             res['network']['name'])
            self.assertFalse(res['network'][qinq_apidef.QINQ_FIELD])

    def _test_create_network_qinq_and_transparent_vlan(self, qinq_value, vlt):
        arg_list = (
            qinq_apidef.QINQ_FIELD,
            vlan_apidef.VLANTRANSPARENT)
        net_kwargs = {
            qinq_apidef.QINQ_FIELD: qinq_value,
            vlan_apidef.VLANTRANSPARENT: vlt,
        }
        # Both vlan_transparent and qinq can't be set for the same network
        if qinq_value and vlt:
            with testlib_api.ExpectedException(
                    web_exc.HTTPClientError) as ctx_manager:
                with self.network(name='net1', as_admin=True,
                                arg_list=arg_list, **net_kwargs):
                    pass
            self.assertEqual(web_exc.HTTPBadRequest.code,
                            ctx_manager.exception.code)
            return

        # In any other case it should work fine
        with self.network(name='net1', as_admin=True,
                          arg_list=arg_list, **net_kwargs) as net:
            req = self.new_show_request('networks', net['network']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertEqual(net['network']['name'],
                             res['network']['name'])
            self.assertEqual(qinq_value,
                             res['network'][qinq_apidef.QINQ_FIELD])
            self.assertEqual(vlt, res['network'][vlan_apidef.VLANTRANSPARENT])

    def test_create_network_qinq_disabled_transparent_vlan_enabled(self):
        self._test_create_network_qinq_and_transparent_vlan(False, True)

    def test_create_network_qinq_disabled_transparent_vlan_disabled(self):
        self._test_create_network_qinq_and_transparent_vlan(False, False)

    def test_create_network_qinq_enabled_transparent_vlan_disabled(self):
        self._test_create_network_qinq_and_transparent_vlan(True, False)

    def test_create_network_qinq_enabled_transparent_vlan_enabled(self):
        self._test_create_network_qinq_and_transparent_vlan(True, True)



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import functools

from alembic import context
from alembic import op
import sqlalchemy as sa

from neutron._i18n import _

# Neutron milestones for upgrade aliases
LIBERTY = 'liberty'
MITAKA = 'mitaka'
NEWTON = 'newton'
OCATA = 'ocata'
PIKE = 'pike'
QUEENS = 'queens'
ROCKY = 'rocky'
STEIN = 'stein'
TRAIN = 'train'
USSURI = 'ussuri'
VICTORIA = 'victoria'
WALLABY = 'wallaby'
XENA = 'xena'
YOGA = 'yoga'
ZED = 'zed'
RELEASE_2023_1 = '2023.1'
RELEASE_2023_2 = '2023.2'
RELEASE_2024_1 = '2024.1'
RELEASE_2024_2 = '2024.2'
RELEASE_2025_1 = '2025.1'

NEUTRON_MILESTONES = [
    # earlier milestones were not tagged
    LIBERTY,
    MITAKA,
    NEWTON,
    OCATA,
    PIKE,
    QUEENS,
    ROCKY,
    STEIN,
    TRAIN,
    USSURI,
    VICTORIA,
    WALLABY,
    XENA,
    YOGA,
    ZED,
    RELEASE_2023_1,
    RELEASE_2023_2,
    RELEASE_2024_1,
    RELEASE_2024_2,
    # Do not add the milestone until the end of the release
]


def skip_if_offline(func):
    """Decorator for skipping migrations in offline mode."""
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        if context.is_offline_mode():
            return
        return func(*args, **kwargs)

    return decorator


def raise_if_offline(func):
    """Decorator for raising if a function is called in offline mode."""
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        if context.is_offline_mode():
            raise RuntimeError(_("%s cannot be called while in offline mode") %
                               func.__name__)
        return func(*args, **kwargs)

    return decorator


@raise_if_offline
def schema_has_table(table_name):
    """Check whether the specified table exists in the current schema.

    This method cannot be executed in offline mode.
    """
    insp = sa.inspect(op.get_bind())
    return table_name in insp.get_table_names()


@raise_if_offline
def schema_has_column(table_name, column_name):
    """Check whether the specified column exists in the current schema.

    This method cannot be executed in offline mode.
    """
    insp = sa.inspect(op.get_bind())
    # first check that the table exists
    if not schema_has_table(table_name):
        return
    # check whether column_name exists in table columns
    return column_name in [column['name'] for column in
                           insp.get_columns(table_name)]


@raise_if_offline
def alter_column_if_exists(table_name, column_name, **kwargs):
    """Alter a column only if it exists in the schema."""
    if schema_has_column(table_name, column_name):
        op.alter_column(table_name, column_name, **kwargs)


@raise_if_offline
def drop_table_if_exists(table_name):
    if schema_has_table(table_name):
        op.drop_table(table_name)


@raise_if_offline
def rename_table_if_exists(old_table_name, new_table_name):
    if schema_has_table(old_table_name):
        op.rename_table(old_table_name, new_table_name)


def alter_enum_add_value(table, column, enum, nullable, server_default=None):
    '''Expand Enum values for a column.

    :param table:table name
    :param column: column name
    :param enum: sqlalchemy Enum with updated values
    :param nullable: existing nullable for column.
    :param server_default: existing or new server_default for the column
    '''
    op.alter_column(table, column, type_=enum,
                    existing_nullable=nullable,
                    server_default=server_default)


def get_unique_constraints_map(table):
    inspector = sa.inspect(op.get_bind())
    return {
        tuple(sorted(cons['column_names'])): cons['name']
        for cons in inspector.get_unique_constraints(table)
    }


def remove_fk_unique_constraints(table, foreign_keys):
    unique_constraints_map = get_unique_constraints_map(table)
    for fk in foreign_keys:
        constraint_name = unique_constraints_map.get(
            tuple(sorted(fk['constrained_columns'])))
        if constraint_name:
            op.drop_constraint(
                constraint_name=constraint_name,
                table_name=table,
                type_="unique"
            )


def remove_foreign_keys(table, foreign_keys):
    for fk in foreign_keys:
        op.drop_constraint(
            constraint_name=fk['name'],
            table_name=table,
            type_='foreignkey'
        )


def create_foreign_keys(table, foreign_keys):
    for fk in foreign_keys:
        op.create_foreign_key(
            constraint_name=fk['name'],
            source_table=table,
            referent_table=fk['referred_table'],
            local_cols=fk['constrained_columns'],
            remote_cols=fk['referred_columns'],
            ondelete=fk['options'].get('ondelete')
        )


@contextlib.contextmanager
def remove_fks_from_table(table, remove_unique_constraints=False):
    try:
        inspector = sa.inspect(op.get_bind())
        foreign_keys = inspector.get_foreign_keys(table)
        remove_foreign_keys(table, foreign_keys)
        if remove_unique_constraints:
            remove_fk_unique_constraints(table, foreign_keys)
        yield
    finally:
        create_foreign_keys(table, foreign_keys)


def pk_on_alembic_version_table():
    inspector = sa.inspect(op.get_bind())
    pk = inspector.get_pk_constraint('alembic_version')
    if not pk['constrained_columns']:
        op.create_primary_key(op.f('pk_alembic_version'),
                              'alembic_version', ['version_num'])
